"""Base handler classes."""
import abc
import atexit
import json
import logging
import tempfile
from hashlib import md5
from os.path import expanduser
from pathlib import Path
from typing import Any, Dict, Iterator, Optional, Union

import requests

log = logging.getLogger(__name__)


class IndexHandler(metaclass=abc.ABCMeta):
    """Creates a base index handler to be extended."""

    @abc.abstractmethod
    def __init__(self, path: str, cache_dir: Optional[str] = None) -> None:
        """Ensure a string path is provided."""
        ...

    @abc.abstractmethod
    def __getitem__(self, index) -> Any:
        """Add the ability to extend __getitem__."""
        ...

    @abc.abstractmethod
    def __len__(self) -> int:
        """Add the ability to extend __len__."""
        ...

    @abc.abstractmethod
    def __iter__(self) -> Iterator:
        """Add the ability to extend __iter__."""
        ...

    @abc.abstractmethod
    def packages(self) -> Dict:
        """Add the ability to extend __getitem__."""
        ...

    @abc.abstractmethod
    def find(self) -> Union[list, None]:
        """Add interface to index finder."""
        ...

    @abc.abstractmethod
    def update(self) -> bool:
        """Add interface for extendable updater."""
        ...

    @staticmethod
    @abc.abstractmethod
    def handles(index: Union[str, Path]) -> bool:
        """Add interface for checking if a handler can handle index."""
        ...

    @property
    @abc.abstractmethod
    def outdated(self) -> bool:
        """Add interface to check if an index is outdated."""
        ...


class JsonFileIndexHandler(IndexHandler):
    """Local file handler."""

    def __init__(self, path: str, cache_dir: Optional[str] = None) -> None:
        """With the given path, process data and return the results."""
        path = str(path)[7:] if str(path)[:7] == 'file://' else str(path)
        self._path = Path(expanduser(path)).resolve()
        if not self._path.exists():
            message = 'index file not found'
            raise FileNotFoundError(message)
        with open(self._path) as json_file:
            self._data = json.load(json_file)

    def __getitem__(self, index) -> Any:
        """Get an item from the _data list read from the json file."""
        return self._data.__getitem__(index)

    def __len__(self) -> int:
        """Return length of _data list returned from the json file."""
        return self._data.__len__()

    def __iter__(self) -> Iterator:
        """Return the list of packages contained within the index."""
        return iter(self._data)

    def packages(self) -> Dict:
        """Return the list of packages defined."""
        return self._data

    def find(self, value: str) -> Union[list, None]:
        """Find packages matching a specified value."""
        return [s for s in self._data if value.lower() in s.lower()]

    def update(self) -> bool:
        """Update the index file."""
        return True

    @staticmethod
    def handles(index: Union[str, Path]) -> bool:
        """Check if the index format is file:// or a path."""
        path = str(index)[7:] if str(index)[:7] == 'file://' else index
        try:
            return Path(str(path)).exists()
        except OSError:
            return False

    @property
    def outdated(self) -> bool:
        """Check if the index is outdated."""
        return False


class JsonUrlIndexHandler(IndexHandler):
    """Use a json file located on a remote server."""

    def __init__(self, index: str, cache_dir: Optional[str] = None) -> None:
        """With given path, process the data and return the results."""
        self._index = index
        if not cache_dir:
            tmp_cache_dir = tempfile.TemporaryDirectory()
            atexit.register(tmp_cache_dir.cleanup)
            cache_dir = str(tmp_cache_dir)
        self._cache = Path(cache_dir)
        self._cache.mkdir(0o777, parents=True, exist_ok=True)
        log.info(f'Creating dir {self._cache}')
        self._cached_index = Path(self._cache, 'index.json')
        self._cached_index.touch(exist_ok=True)
        self._updated = False
        self.update()
        with open(self._cached_index) as json_file:
            self._data = json.load(json_file)

    def __getitem__(self, index) -> Any:
        """Get an item from the _data list read from the json file."""
        return self._data.__getitem__(index)

    def __len__(self) -> int:
        """Return length of _data list returned from the json file."""
        return self._data.__len__()

    def __iter__(self) -> Iterator:
        """Return the list of packages contained within the index."""
        return iter(self._data)

    def packages(self) -> Dict:
        """Return the list of packages defined."""
        return self._data

    def find(self, value: str) -> Union[list, None]:
        """Find packages matching a specified value."""
        return [s for s in self._data if value.lower() in s.lower()]

    def update(self) -> None:
        """Update the index file if its outdated."""
        self._updated = False
        headers = {'If-None-Match': self._digest}
        req = requests.get(self._index, headers=headers, allow_redirects=True)
        if req.status_code not in [200, 304]:
            raise FileNotFoundError(req.status_code)
        elif req.status_code == 200:
            with open(self._cached_index, 'wb') as cached_index:
                cached_index.write(req.content)
            self._updated = True

    @staticmethod
    def handles(index: Union[str, Path]) -> bool:
        """Check if the index format is file:// or a path."""
        return str(index)[0:4] == 'http'

    @property
    def outdated(self) -> bool:
        """Check if an index is outdated.

        Execute a head request using the requests library to check that
        the ETag matches.
        """
        headers = {'If-None-Match': self._digest}
        req = requests.head(self._index, headers=headers, allow_redirects=True)
        if req.status_code not in [200, 304]:
            raise FileNotFoundError(req.status_code)
        elif req.status_code == 200:
            return True
        else:
            return False

    @property
    def _digest(self) -> str:
        """Return the md5 digest of the currently cached index file."""
        digest = md5()
        with open(self._cached_index, 'rb') as cached_index:
            for block in iter(lambda: cached_index.read(65536), b''):
                digest.update(block)
        return digest.hexdigest()



"""Dismantle is a Python package / plugin / extension manager."""
import logging

from dismantle.__version__ import __version__  # noqa: F401

logging.basicConfig(
    level='DEBUG',
    format='%(asctime)s - %(name)s - [ %(message)s ]',
    datefmt='%d-%b-%y %H:%M:%S',
    force=True,
    handlers=[logging.NullHandler()],
)
log = logging.getLogger(__name__)



"""Plugin management."""
_plugins = {}

from dismantle.plugin.iplugin import IPlugin  # noqa: E402
from dismantle.plugin.plugin import plugin  # noqa: E402
from dismantle.plugin.plugins import Plugins  # noqa: E402
from dismantle.plugin.register import register  # noqa: E402

__all__ = [
    '_plugins',
    'IPlugin',
    'register',
    'plugin',
    'Plugins'
]

# IDEAS: --- plugins ----------------------------
# IDEAS: Func hook to provide plugin (process content before / after)
# IDEAS: List plugins for a function



"""Creates a solution to handle multiple package formats."""
from dismantle.package._formats import (
    DirectoryPackageFormat,
    PackageFormat,
    TarPackageFormat,
    TgzPackageFormat,
    ZipPackageFormat
)
from dismantle.package._handlers import (
    HttpPackageHandler,
    LocalPackageHandler,
    PackageHandler
)

__all__ = [
    'HttpPackageHandler',
    'PackageFormat',
    'PackageHandler',
    'DirectoryPackageFormat',
    'LocalPackageHandler',
    'ZipPackageFormat',
    'TarPackageFormat',
    'TgzPackageFormat'
]



from tests.GreetingExtension import GreetingExtension


class AfternoonColorExtension(GreetingExtension):
    _name = 'afternoon'

    def greet(self, person: str) -> None:
        print(f'{self._name} {person}')



"""Provide the ability to extend a module.

Extensions provide the abiity to add additional functionality to a
defined module by providing an interface that can be extended by
multiple classes each with a new provider.
"""

# TODO: activate, deactivate, active

__all__ = [
    'IExtension',
    'Extensions'
]

from dismantle.extension.extensions import Extensions
from dismantle.extension.iextension import IExtension



"""Plugins locator."""
import logging
import os
from importlib.abc import Loader

from dismantle.plugin import IPlugin, _plugins

# enable logging
log = logging.getLogger(__name__)


class Plugins:
    """Search through the installed packages and find plugins."""

    def __init__(self, types, packages, prefix):
        """Initialise plugin system."""
        self._packages = packages
        self._plugins = _plugins
        self._directory = 'plugins'
        self._prefix = prefix
        self._imports = {}
        self._exclude = {'__pycache__'}

        # check that the types are a subclass of IPlugin
        if not all([issubclass(i, IPlugin) for i in types]):
            raise ValueError('all plugin types must extend IPlugin')

        self._types = types
        self._find()

    def _find(self):
        """Search through the packages and find all plugins."""
        for x in self._packages.enabled.values():
            _prefix = x['name']
            # check if the package has an init file
            try:
                root, paths, filenames = next(os.walk(x['directory']))
            except KeyError:
                continue
            # check if we have an extensions directory
            if self._directory in paths:
                plg_files = os.path.join(root, self._directory)
                for root, dirs, files in os.walk(plg_files, topdown=True):
                    dirs[:] = [d for d in dirs if d not in self._exclude]

                    for name in files:
                        # get the path from the beginning of the module
                        fd = os.path.splitext(name)[0]
                        full_path = os.path.join(root, fd)
                        v = full_path[len(plg_files):].replace(os.sep, '.')
                        prefix = f'{_prefix}.plugin{v}'
                        self._imports[prefix] = self._import(full_path, prefix)

    def _import(self, path, prefix):
        """Import a file or module into our imports list."""
        if os.path.isdir(path):
            spec = ('py', 'r', Loader.PKG_DIRECTORY)
            module = Loader.load_module(prefix, None, path, spec)
        else:
            with open(path + '.py', 'r') as plugin_file:
                spec = ('py', 'r', Loader.PY_SOURCE)
                module = Loader.load_module(
                    prefix,
                    plugin_file,
                    f'{path}.py',
                    spec
                )
        return module

    @property
    def plugins(self):
        """Return the list of plugins."""
        return self._plugins

    @property
    def imports(self):
        """Return the list of imported plugins."""
        return self._imports



"""Test packages conpressed as tgz files."""
import os
from pathlib import Path

import pytest

from dismantle.package import PackageFormat, TgzPackageFormat


def test_inherits() -> None:
    assert issubclass(TgzPackageFormat, PackageFormat) is True


def test_grasp_exists(datadir: Path) -> None:
    src = datadir / 'package.tar.gz'
    assert TgzPackageFormat.grasps(src) is True


def test_grasp_file_url(datadir: Path) -> None:
    src = f'file://{datadir.join("package.tgz")}'
    assert TgzPackageFormat.grasps(src) is True


def test_grasp_exists_tgz(datadir: Path) -> None:
    src = datadir / 'package.tgz'
    assert TgzPackageFormat.grasps(src) is True


def test_grasp_not_supported(datadir: Path) -> None:
    src = datadir / 'directory_src'
    assert TgzPackageFormat.grasps(src) is False


def test_grasp_not_supported_format(datadir: Path) -> None:
    src = datadir / 'invalid.file'
    assert TgzPackageFormat.grasps(src) is False


def test_extract_not_supported(datadir: Path) -> None:
    src = datadir / 'directory_src'
    dest = datadir / f'{src}_output'
    message = 'formatter only supports tar.gz files'
    with pytest.raises(ValueError, match=message):
        TgzPackageFormat.extract(src, dest)


def test_extract_not_supported_format(datadir: Path) -> None:
    src = datadir / 'invalid.file'
    dest = datadir / f'{src}_output'
    message = 'formatter only supports tar.gz files'
    with pytest.raises(ValueError, match=message):
        TgzPackageFormat.extract(src, dest)


def test_extract_non_existant(datadir: Path) -> None:
    src = datadir / 'non_existant.tar.gz'
    dest = datadir / f'{src}_output'
    message = 'invalid tgz file'
    with pytest.raises(ValueError, match=message):
        TgzPackageFormat.extract(src, dest)


def test_extract_non_existant_tgz(datadir: Path) -> None:
    src = datadir / 'non_existant.tgz'
    dest = datadir / f'{src}_output'
    message = 'invalid tgz file'
    with pytest.raises(ValueError, match=message):
        TgzPackageFormat.extract(src, dest)


def test_extract_already_exists(datadir: Path) -> None:
    src = datadir / 'package.tar.gz'
    dest = datadir / 'directory_exists'
    assert TgzPackageFormat.extract(src, dest) is None


def test_extract_already_exists_tgz(datadir: Path) -> None:
    src = datadir / 'package.tgz'
    dest = datadir / 'directory_exists'
    assert TgzPackageFormat.extract(src, dest) is None


def test_extract_create(datadir: Path) -> None:
    src = datadir / 'package.tar.gz'
    dest = datadir / 'directory_created'
    TgzPackageFormat.extract(src, dest)
    assert os.path.exists(dest) is True
    assert os.path.exists(dest / 'package.json') is True


def test_extract_create_tgz(datadir: Path) -> None:
    src = datadir / 'package.tar.gz'
    dest = datadir / 'directory_created'
    TgzPackageFormat.extract(src, dest)
    assert os.path.exists(dest) is True
    assert os.path.exists(dest / 'package.json') is True



"""Provides the PackageHandler interface and two example handlers.

Each package can be located in one of multiple locations. Packages can
be located on a remote server, on the local filesystem, or any  other
custom location.  In order to add a new custom location, the
PackageHandler needs to be extended to handle the new custom location.

Currently, two handlers are provided, one to handle packages stored on
the local filesystem, and one to support packages located online and
retrieved using a web request.
"""
import abc
import atexit
import json
import logging
import shutil
import tempfile
from hashlib import md5
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import List, Optional, Union
from urllib.parse import urlparse

import requests

from dismantle.package._formats import (
    DirectoryPackageFormat,
    PackageFormat,
    ZipPackageFormat
)

log = logging.getLogger(__name__)

Formats = Optional[List[PackageFormat]]


class PackageHandler(metaclass=abc.ABCMeta):
    """Base PackageHandler interface.

    The PackageHandler interface defines the structure expected for all
    package handlers and should be extended to provide additional
    custom package handler.
    """

    @abc.abstractmethod
    def __getattr__(self, name) -> object:
        """Allow the __getattr__ method to be extended."""
        ...

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Return a string containing the name of the package."""
        ...

    @property
    @abc.abstractmethod
    def installed(self) -> bool:
        """Return a boolean if the package has been installed."""
        ...

    @staticmethod
    @abc.abstractmethod
    def grasps(path: Union[str, Path]) -> bool:
        """Check if package handler understand a package format."""
        ...

    @abc.abstractmethod
    def install(
        self,
        path: Union[str, Path],
        version: Optional[str] = None
    ) -> bool:
        """Install a specific package version."""
        ...

    @abc.abstractmethod
    def uninstall(self) -> bool:
        """Uninstall a package."""
        ...

    @abc.abstractmethod
    def verify(self, signature: str) -> bool:
        """Verify a packages hash with the provided signature."""
        ...


class LocalPackageHandler(PackageHandler):
    """Directory package structure."""

    def __init__(
        self,
        name: str,
        src: Union[str, Path],
        formats: Optional[Formats] = None
    ) -> None:
        """Initialise the package."""
        self._meta = {}
        self._meta['name'] = name
        self._path = None
        self._installed = False
        self._src = str(src)[7:] if str(src)[:7] == 'file://' else src
        if formats is None:
            formats = [DirectoryPackageFormat]
        for current_format in formats:
            if current_format.grasps(self._src):
                self._format = current_format
                break
        else:
            message = 'unable to process source format'
            raise FileNotFoundError(message)

    @property
    def name(self) -> str:
        """Return the name of the package from the meta data."""
        return self.meta['name']

    @property
    def path(self) -> str:
        """Return the path the package is installed into."""
        return str(self._path)

    def __getattr__(self, name):
        """Return metadata.

        Return an attribute from the meta data if the data doesnt exist.
        """
        if name not in self._meta:
            message = f'{name} is an invalid attribute'
            raise AttributeError(message)
        return self._meta[name]

    @property
    def installed(self) -> bool:
        """Return the current installation state."""
        return self._installed

    @staticmethod
    def grasps(path: Union[str, Path]) -> bool:
        """Check if the package format can process.

        Check if a directory on the local filesystem has been provided.
        """
        path = str(path)[7:] if str(path)[:7] == 'file://' else path
        try:
            return Path(str(path)).exists()
        except OSError:
            return False

    def install(
        self,
        path: Optional[str] = None,
        version: Optional[str] = None
    ) -> bool:
        """Install a package.

        The local package handler does not install the package. No
        version control exists for the directory package type.
        """
        path = str(path)[7:] if str(path)[:7] == 'file://' else path
        self._path = path if path else self._src
        self._format.extract(self._src, self._path)
        self._meta = {**self._meta, **self._load_metadata(self._path)}
        self._installed = True
        return True

    def uninstall(self) -> bool:
        """Uninstall the package."""
        if self._path != self._src:
            self._remove_files(Path(self._path or ''))
        self._path = None
        self._installed = False
        return True

    def verify(self, digest: Optional[str] = None) -> bool:
        """Verify the package hasn't been tampered with."""
        if digest is None:
            return True
        message = 'the local package handler does not support verification'
        raise ValueError(message)

    def _load_metadata(self, path: Union[str, Path]):
        """Load the package.json file into memory."""
        path = Path(str(path)[7:] if str(path)[:7] == 'file://' else path)
        try:
            with open(path / 'package.json') as package:
                meta = json.load(package)
                if 'name' not in meta:
                    message = 'meta file missing name value'
                    raise ValueError(message)
                if self._meta['name'] != meta['name']:
                    message = 'meta name does not match provided package name'
                    raise ValueError(message)
                if 'version' not in meta:
                    message = 'meta file missing version value'
                    raise ValueError(message)
                return meta
        except JSONDecodeError:
            message = 'invalud package file format'
            raise ValueError(message)

    @staticmethod
    def _remove_files(path: Union[str, Path]) -> None:
        """Recursively remove the path and all its sub items."""
        path = str(path)[7:] if str(path)[:7] == 'file://' else path
        try:
            shutil.rmtree(path)
        except OSError:
            FileNotFoundError('unable to remove files')


class HttpPackageHandler(PackageHandler):
    """Url package structure."""

    def __init__(
        self,
        name: str,
        src: Union[str, Path],
        formats: Formats = None
    ):
        """Initialise the package."""
        self._meta = {}
        self._meta['name'] = name
        self._path = None
        self._installed = False
        self._updated = False
        self._src = src

        tmp_cache = tempfile.TemporaryDirectory()
        cache_dir = Path(tmp_cache.name)
        atexit.register(tmp_cache.cleanup)

        parts = urlparse(str(src))
        ext = ''.join(Path(parts.path).suffixes)
        self._cache = Path(cache_dir / Path(name + ext))
        if not HttpPackageHandler.grasps(src):
            message = 'invalid handler format'
            raise ValueError(message)
        if formats is None:
            formats = [ZipPackageFormat]
        for current_format in formats:
            if current_format.grasps(self._src):
                self._format = current_format
                break
        else:
            message = 'a valid source is required'
            raise FileNotFoundError(message)

    @property
    def name(self) -> str:
        """Return the name of the package from the meta data."""
        return self.meta['name']

    def __getattr__(self, name):
        """Return attrib from meta data if the data doesnt exist."""
        if name not in self._meta:
            message = f'{name} is an invalid attribute'
            raise AttributeError(message)
        return self._meta[name]

    @property
    def installed(self) -> bool:
        """Return the current installation state."""
        return self._installed

    @staticmethod
    def grasps(path: Union[str, Path]) -> bool:
        """Check if dir on the local filesystem has been provided."""
        parts = urlparse(str(path))
        if parts.scheme not in ['http', 'https']:
            return False
        return True

    def _fetch_and_extract(self):
        headers = {'If-None-Match': self._digest}
        req = requests.get(
            str(self._src),
            headers=headers,
            allow_redirects=True
        )
        if req.status_code not in [200, 304]:
            raise FileNotFoundError(req.status_code)
        elif req.status_code == 200:
            # Make sure parent folder of the cache exists
            self._cache.parents[0].mkdir(0o777, parents=True, exist_ok=True)
            log.info(f'Creating dir {self._cache.parents[0]} to hold cache')
            # Write the package raw bytes into a single cache file
            with open(self._cache, 'wb') as cached_package:
                cached_package.write(req.content)
            self._updated = True
        self._format.extract(self._cache, self._path or '')

    def install(self, path: str, version: Optional[str] = None) -> bool:
        """Install the current package to the given path.

        If there's already a package in path we'll only fetch if the
        version is different.
        """
        fetch_required = True
        try:
            existing_pkg_metadata = self._load_metadata(Path(path))
            if existing_pkg_metadata['version'] == self._meta['version']:
                fetch_required = False
        except ValueError:
            # Ignore _load_metadata errors
            pass
        except OSError:
            # Ignore Not Found
            pass
        except KeyError:
            # ignore if `_meta` is empty
            pass

        self._path = path
        self._updated = False

        if fetch_required:
            self._fetch_and_extract()

        self._meta = {**self._meta, **self._load_metadata(Path(self._path))}
        self._installed = True
        return True

    def uninstall(self) -> bool:
        """Uninstall the package."""
        if self._path != self._src:
            self._remove_files(Path(self._path or ''))
        self._path = None
        self._installed = False
        return True

    def verify(self, digest: Optional[str] = None) -> bool:
        """Verify the package hasn't been tampered with."""
        if digest is None:
            return True
        message = 'the http package handler does not support verification'
        raise ValueError(message)

    def _load_metadata(self, path: Path):
        """Load the package.json file into memory."""
        try:
            with open(path / 'package.json') as package:
                meta = json.load(package)
                if 'name' not in meta:
                    message = 'meta file missing name value'
                    raise ValueError(message)
                if self._meta['name'] != meta['name']:
                    message = 'meta name does not match provided package name'
                    raise ValueError(message)
                if 'version' not in meta:
                    message = 'meta file missing version value'
                    raise ValueError(message)
                return meta
        except JSONDecodeError:
            message = 'invalid package file format'
            raise ValueError(message)

    @staticmethod
    def _remove_files(path: Path) -> None:
        """Recursively remove the path and all its sub items."""
        try:
            shutil.rmtree(path)
        except OSError:
            FileNotFoundError('unable to remove files')

    @property
    def outdated(self) -> bool:
        """Execute a head request using the requests library.

        To check that the ETag matches.
        """
        headers = {'If-None-Match': self._digest}
        req = requests.head(
            str(self._src or ''),
            headers=headers,
            allow_redirects=True
        )

        if req.status_code not in [200, 304]:
            raise FileNotFoundError(req.status_code)
        elif req.status_code == 200:
            return True
        else:
            return False

    @property
    def _digest(self) -> str:
        """Return the md5 digest of the currently cached index file."""
        digest = md5()
        if not self._cache.exists():
            return digest.hexdigest()
        with open(self._cache, 'rb') as cached_package:
            for block in iter(lambda: cached_package.read(65536), b''):
                digest.update(block)
        return digest.hexdigest()



"""Provide the ability to decorate functions with a plugin decorator."""
import logging

from dismantle.plugin import _plugins

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class plugin:  # noqa: N801
    """Plugin decorator to register plugins."""

    def __init__(self, name, priority=100):
        """Get the name of the registered method.

        To add the plugin to and set the evecution priority. Priority
        less than zero executes before the main method, priority above
        zero executes afterwards.
        """
        log.debug(f'plugin.__init__: ({priority}) {name}')

        # ensure the priority is numeric
        if not isinstance(priority, int):
            raise ValueError('Priority needs to be numeric')
        self._priority = priority

        # check the name is printable
        if not name.isprintable():
            raise ValueError('Name needs to be printable')

        # check the plugin registration exists
        self._name = name
        if name not in _plugins:
            log.warning(f'function "{name}" not registered')
            _plugins[name] = {'original': None}
            _plugins[name]['plugins'] = []

    def __call__(self, extend_function):
        """Call wrapper to be executed when a function is wrapped."""
        log.debug(f'plugin.__call__: {extend_function}')
        self._function = extend_function
        global _plugins
        _plugins[self._name]['plugins'].append({
            'function': self._function,
            'priority': self._priority
        })



"""Provides the ability to handle package index files."""
from typing import List, Type

from dismantle.index._handlers import (
    IndexHandler,
    JsonFileIndexHandler,
    JsonUrlIndexHandler
)

__all__ = [
    'IndexHandler',
    'JsonFileIndexHandler',
    'JsonUrlIndexHandler'
]

indicies_list: List[str] = []
handlers_list: List[Type[IndexHandler]] = []
cache: str


def get_packages():
    """Get the list of package meta from all the provided indexes."""
    packages = {}

    for index in indicies_list:
        for handler in handlers_list:
            if handler.handles(index):
                data = handler(index, cache)
                packages.update(data.packages())

    return packages


def add_indicies(indicies: List[str]) -> None:
    """Add the indicies to be processed."""
    global indicies_list
    indicies_list.extend(indicies)


def set_cache(cache_dir: str) -> None:
    """Set the cache directory to store the index."""
    global cache
    cache = cache_dir


def add_handlers(handlers: List[Type[IndexHandler]]) -> None:
    """Add a handler to the list of supported index handlers."""
    global handlers_list
    handlers_list.extend(handlers)



"""Provides an extension to be used for testing."""
from dismantle.extension import IExtension


class ColorExtension(IExtension):
    """An example extension defining a set of colors."""

    _category = 'color'

    def color(self) -> None:
        """Set a color for the test extension."""
        ...



"""Register plugin hooks."""
import functools
import logging
from logging import NullHandler

from dismantle.plugin import _plugins

log = logging.getLogger(__name__).addHandler(NullHandler())


def register(name=None):
    """Register a function to add plugins to."""
    if not name:
        raise ValueError

    def decorator(_func: functools._AnyCallable):
        _plugins[name] = {'original': _func}
        _plugins[name]['plugins'] = []

        @functools.wraps(_func)
        def caller(self, *args, **kwargs):
            """Execute the registered plugin in order of priority."""
            # store the previous result
            _plugins[name]['result'] = args[0]

            # get the list of before and after plugins
            before = [p for p in _plugins[name]['plugins'] if p['priority'] < 0]  # NOQA: E501
            after = [p for p in _plugins[name]['plugins'] if p['priority'] >= 0]  # NOQA: E501

            # execute all the before plugins
            for _name in before:
                _args = (self, _plugins[name]['result'])
                _plugins[name]['result'] = _name['function'](*_args)

            # execute the actual method that was registered
            _args = (self, _plugins[name]['result'])
            _plugins[name]['result'] = _plugins[name]['original'](*_args)

            # execute all the after plugins
            for _name in after:
                _args = (self, _plugins[name]['result'])
                _plugins[name]['result'] = _name['function'](*_args)

            return _plugins[name]['result']
        return caller
    return decorator
